var fs			= require('fs');
var fs_extra 	= require('fs-extra')
var path		= require('path');
var _			= require('underscore');
var inquirer	= require('inquirer');
var request		= require('request');
var open		= require('open');
var archiver 	= require('archiver');
var semver		= require('semver');
var Q 			= require('q');
var shelljs 	= require('shelljs');
var Multibar 	= require('./multibar');
var unzip		= require('unzip2');
var extract		= require('extract-zip');

var create = function(appPath) {
	// init project at given path or default to cwd
	appPath = (typeof appPath == 'string') ? appPath : process.cwd();
	
	// check if chosen path exists and is a directory
	if( !fs.existsSync(appPath) ) return console.error("Error: path does not exist");
	if( !fs.lstatSync(appPath).isDirectory() ) return console.error("Error: path is not a directory");
	
	// ask some questions to setup the project
	inquirer.prompt([
		{
			type: "input",
			name: "id",
			message: "What is your app ID? (can be found in the developers center)",
		},
		{
			type: "name",
			name: "name",
			message: "What is the name of your app?",
			default: "myFormideApp"
		},
		{
			type: "description",
			name: "description",
			message: "What is the description of your app?",
			default: "This is an awesome app!"
		},
		{
			type: "input",
			name: "version",
			message: "What is you app version? (fill in a semver compatible version number)",
			default: "1.0.0"
		},
		{
			type: "template",
			name: "template",
			message: "Which starter template? (leave empty or choose from: 'basic')"
		},
		{
			type: "confirm",
			name: "confirm",
			message: "Use this config?"
		}
	], function(input) {
		var projectPath = path.join(appPath, input.name);
		
		// check if version valid
		if(semver.valid(input.version) == null) {
			return console.log("Version number invalid, make sure you use a valid semver version number".red);
		}
		
		// check if dir already exists
		if (fs.existsSync(projectPath)) return console.log(("Error: path " + projectPath + " already exist").red);
		
		// create project dir
		fs.mkdirSync(projectPath);
		
		// build manifest
		var manifest = {
			id: input.id,
			name: input.name,
			version: input.version
		};
		
		// add author when logged in
		if (typeof lib.settings.user == 'object') {
			manifest.author = {
				name: lib.settings.user.firstname + " " + lib.settings.user.lastname,
				email: lib.settings.user.email	
			};
		}
		
		// create app.json
		fs.writeFileSync(path.join(projectPath, 'app.json'), JSON.stringify(manifest, null, 4));
		
		// create README.md
		fs.writeFileSync(path.join(projectPath, 'README.md'), "# " + input.name + "\n" +input.description);
		
		// copy chosen template
		if(input.template) {
			return getGithubTemplate(input.template, projectPath);
		}
	});
};

var getGithubTemplate = function(templateName, targetPath) {
	var q = Q.defer();
	var repoPath = "https://github.com/PRINTR3D/formide-apptemplate-" + templateName;
	var repoFolderPath = "archive";
	
	// select latest master zip
	repoPath += "/archive/master.zip";
	
	getArchive(repoPath, targetPath + '/' + repoFolderPath)
	.then(function() {
		try {
			// move the template contents to the www folder
			cp('-Rf', targetPath + '/' + repoFolderPath + '/.', targetPath + '/www');
			
			// clean this folder
			rm('-rf', tagetPath + '/' + repoFolderPath + '/');
			
			q.resolve();
		}
		catch (e) {
			q.reject(e);
		}
	})
	.catch(function(err) {
		console.error('Make sure you enter a valid template name');
	});
	
	return q.promise;
}

var info = function() {
	// run project at given path or default to cwd
	appPath = (typeof appPath == 'string') ? appPath : process.cwd();
	
	// get manifest contents for upload info
	var manifest = fs.readFileSync(appPath + '/app.json', 'utf8');
	manifest = JSON.parse(manifest);
	
	// show manifest contents
	console.log(manifest);
	
	// wait a bit and exit
	setTimeout(process.exit, 1000);
};

var run = function() {
	// run project at given path or default to cwd
	appPath = (typeof appPath == 'string') ? appPath : process.cwd();
	
	// open local app index.html in browser
	open('file://' + appPath + '/index.html');
	
	// wait a bit and exit
	setTimeout(process.exit, 1000);
};

var publish = function(appPath) {
	// publish project at given path or default to cwd
	appPath = (typeof appPath == 'string') ? appPath : process.cwd();
	
	// get manifest contents for upload info
	var manifest = fs.readFileSync(appPath + '/app.json', 'utf8');
	manifest = JSON.parse(manifest);
	
	// compress and upload app
	compress(appPath, manifest, function(err, zipPath) {
		upload(zipPath, manifest, function(err, httpResponse, body) {
			
			// remove zipfile
			fs.unlink(zipPath);
			
			// handle upload error
			if (err) {
				return console.log(('Upload failed: ' + err.message).red);
			}
			
			try {
				var response = JSON.parse(body);
				
				// handle not 200 OK response from api as error
				if (httpResponse.statusCode !== 200) {
					return console.log(('Upload failed: ' + response.message).red);
				}
				
				// patch bump manifest version number
				manifest.version = semver.inc(manifest.version, 'patch');
				fs.writeFileSync(appPath + '/app.json', JSON.stringify(manifest), 'utf8');
				
				// log success
				console.log(('Upload complete: ' + response.message).green);
				
				// wait a bit and exit
				setTimeout(process.exit, 1000);
			}
			catch (e) {
				return console.log(('Upload failed: ' + e.message).red);
			}
		});
	});
};

var compress = function(appPath, manifest, callback) {
	var zipPath = appPath + '/app.zip';
	var zip = fs.createWriteStream(zipPath);
	var archive = archiver('zip');
	
	archive.pipe(zip);
	archive.bulk([{
		expand: true,
		cwd: appPath,
		src: ['**']
	}]);
	
	archive.finalize(function(err, bytes) {
		if(err) {
			console.log(err);
		}
	});
	
	zip.on('close', function() {
		callback(null, zipPath);
	});
};

var upload = function(zipPath, manifest, callback) {
	lib.api.call({
		path: '/apps/upload',
		method: 'POST',
		json: false,
		appUpload: {
			version: manifest.version,
			id: manifest.id,
			zipPath: zipPath
		}
	}, callback);
}; 

var getArchive = function(archiveUrl, targetPath) {
	var q = Q.defer();
	var os = require('os');
	
	console.log('Downoading template from ' + archiveUrl);
	
	var tmpFolder = os.tmpDir();
	var tmpPath = path.join(tmpFolder, 'formide-apptemplate-' + new Date().getTime() + '.zip');
	
	var unzipArchive = function(filename) {
		var readStream = fs.createReadStream(filename);
		var writeStream = unzip.Extract({ path: targetPath });

		readStream.on('error', function(err) {
			console.error('An error occurred while unzipping template archive: ' + err.message);
			q.reject(err);
			return;
		});
		
		writeStream.on('close', function() {
			q.resolve();
		});
		
		writeStream.on('error', function(err) {
			console.error('An error occurred while writing template archive: ' + err.message);
			q.reject(err);
			return;
		});
		
		readStream.pipe(writeStream);
	}
	
	request({
		url: archiveUrl,
		rejectUnauthorized: false,
		encoding: null
	}, function(err, res, body) {
		if (err) {
			console.error('An error occurred while downloading template archive: ' + err.message);
			q.reject(err);
			return;
		}
		
		if (!res) {
			console.error('Invalid response while downloading template archive: ' + archiveUrl);
			q.reject('Unable to parse response');
			return;
		}
		
		if (res.statusCode !== 200) {
			if (res.statusCode === 404 || res.statusCode === 406) {
				console.error('Template archive not found at url ' + archiveUrl);
			}
			else {
				console.error('Invalid response status while downloading template archive: ' + archiveUrl);
			}
			q.reject(res);
			return;
		}
		
		try {
			fs.writeFileSync(tmpPath, body);
			unzipArchive(tmpPath);
		}
		catch(e) {
			console.error('An error occured while unzipping template archive: ' + e.message);
			q.reject(e);
		}
	})
	.on('response', function(res) {
		var bar = Multibar.newBar('[:bar]  :percent  :etas', {
			complete: '=',
			incomplete: ' ',
			width: 30,
			total: parseInt(res.headers['content-length'], 10)
		});
		
		res.on('data', function(chunck) {
			try {
				bar.tick(cunck.length);
			}
			catch(e) {}
		});
	});
	
	return q.promise;
}

module.exports.create = create;
module.exports.publish = publish;
module.exports.run = run;
module.exports.info = info;